program pe;     { simple pascal editor }
{ written by Russell Whitaker
  v 0.1 - Jan 2000
  v 0.2 - Feb 2002 released under GPL
  v 0.3 - Oct 2002 search/replace: y/n/a/q
                   lineinsert: activated
  v 0.4 - Jan 2003 added "w" to search options
  v 0.5 - Feb 2003 changed test for "w"
                   }
uses Crt, GPC, GPCUtil,    {gpc supplied units}
     StringShift, pePrint,
     peBuffer, { 16 procs & functions, starting with BB_ }
     peMenu;   { menu items, dialogs, messages }


const
  MinScreenSizeX : integer = 60;
  MinScreenSizeY : integer = 22;

var
  Ch : Char;                            {temp char}
  CharAttr : TCharAttr;
  i,j   : integer = 0;                  {temp vars}
  inp   : tkey;                         {input key}
  InsertMode : boolean = true;          {insert mode/ overwrite mode toggle}
  state1,
  state2 : winstate;                    {windows}
  Xoffset : integer = 0;                {horiz displacement}
  Yoffset : integer = 0;                {index of first line in display}
  CurLine : lline;                      {Current Line = work space}
  UpdateFlag : boolean = false;         {true = new data in CurLine}
  MarkStart : integer = 0;              {start of marked block}
  MarkEnd : integer = 0;                {end of marked block}
  Pattern : NameString;                 {find string}
  RepPattern : NameString;              {replace string}
  CaseFlag,                             {true = case sensitive}
  ReplaceFlag,                          {true = replace}
  WordFlag : boolean;                   {true = find exact word}
  StartLine,
  EndLine : integer;                    {for search/continue}
  found   : integer = 0;                { search/repl stats }
  replaced: integer = 0;

procedure DisplayLine( Index: integer );
var
 SetColor : TTextAttr;
 Dline : string[ GetXMax ];
 i : integer;
begin
  if (( MarkEnd >= ( Index + Yoffset )) and
      ( MarkStart <= ( Index + Yoffset ))) then
    SetColor := color[ 1 ]
  else
    SetColor := color[ 3 ];
  Dline := copy( BB_ReadLine( Index + Yoffset ), Xoffset + 1, GetXMax );
  WriteStrAt( 1, Index, Dline, SetColor );
  for i := 1 to GetXMax do
  begin
    if i > length( Dline ) then break;
    if  Dline[ i ] = chTab then
      ChangeTextAttr( i, Index, 1, red);
  end;
end;

procedure ResetCurLine;
begin
  FillChar( CurLine[ 1 ], CurLine.capacity, ' ');
  {$X+}
  setlength( CurLine, 0 );
  {$X-}
  WriteCharAt( 1, WhereY, GetXMax, ' ', textattr );
  UpdateFlag := false;
  DisplayLine( WhereY );
end;

function Trimit( Start: integer ): integer;  {trim blanks, return index}
var
  i : integer;
begin
  i := Start;
  while ( i > 1 ) and then
        ( CurLine[ i ] = ' ') or ( CurLine[ i ] = chr( kbTab )) do
    dec( i );
  Trimit := i;
end;

procedure WriteEditWindow;
var
  i, j,
  x, y : integer;
begin
  x := WhereX;
  y := WhereY;
  textbackground( blue );
  ClrScr;
  GoToXY( x, y );
  if ( Yoffset + GetYmax ) < BB_count then              {not EOF}
    j := GetYmax
  else
    j := BB_count - Yoffset;
  for i := 1 to j do
    DisplayLine( i );

  if UpdateFlag then
    for i := 1 to GetXMax do
      if CurLine[ i + Xoffset ] = chTab then
        WriteCharAt( i, WhereY, 1, CurLine[ i + Xoffset ], red )
      else
        WriteCharAt( i, WhereY, 1, CurLine[ i + Xoffset ], textattr );
end;

{------------------------------Cursor Moves------------------------------------}

procedure CursorDown;
var
  i : integer;
begin
  if ( Yoffset + WhereY ) <= BB_count then        {allows cursor at 1st empty line}
    if WhereY < GetYMax then
      GoToXY( WhereX, WhereY + 1 )
    else
    begin
      i := WhereX;
      writeln;
      inc( Yoffset );
      DisplayLine( WhereY );       {GetYmax s/b = WhereY}
      GoToXY( i, WhereY );
    end;
end;

procedure CursorEnd;
var
  i : integer = CurLine.capacity;
begin
  if not UpdateFlag then                         {get line to work on}
  begin
    ResetCurLine;
    CurLine := BB_ReadLine( WhereY + Yoffset );
  end;
  while ( i > 0 ) and ( CurLine[ i ] = ' ') do   {find end}
    dec( i );
  if ( i >= Xoffset ) and ( i < ( GetXMax + Xoffset )) then  {end on screen}
    i := i - Xoffset
  else                                           {new Xoffset needed}
  begin
    if i < GetXMax then                          {low boundry}
      Xoffset := 0
    else
      if i > ( CurLine.capacity - GetXMax ) then   {high boundry}
        Xoffset := CurLine.capacity - GetXMax
      else                                       {mid screen}
        Xoffset := i - ( GetXmax div 2 );
    i := i - Xoffset;
    WriteEditWindow;
  end;
  GoToXY( i + 1, WhereY );
end;

procedure CursorHome;
begin
  if Xoffset > 0 then
  begin
    Xoffset := 0;
    WriteEditWindow;
  end;
  GoToXY( 1, WhereY );
end;

procedure CursorLeft;
begin
  if WhereX > 1 then
    GoToXY( WhereX - 1, WhereY )
  else
    if Xoffset > 0 then
    begin
      dec( Xoffset );
      WriteEditWindow;
      GoToXY( 1, WhereY );
    end;
end;

procedure CursorRight;
begin
  if WhereX < GetXMax then
    GoToXY( WhereX + 1, WhereY )
  else
    if Xoffset < ( CurLine.capacity - GetXMax ) then
    begin
      inc( Xoffset );
      WriteEditWindow;
      GoToXY( GetXMax, WhereY );
    end;
end;

procedure CursorUp;
begin
  ResetCurLine;
  if WhereY > 1 then              { not at top of screen}
    GoToXY( WhereX, WhereY - 1 )
  else
    if Yoffset > 0 then
    begin
      InsLine;
      dec( Yoffset );
      DisplayLine( 1 );
    end;
end;

{--------------------------Additional functions-------------------}

procedure PageDown;
var
  y : integer;
begin
  ResetCurLine;
  if( Yoffset + 20 ) < BB_count then
     Yoffset := Yoffset + 20;
  if ( Yoffset + WhereY ) > BB_count then
    y := BB_count - Yoffset + 1                 {allows go past end}
  else
    y := WhereY;
  WriteEditWindow;
  GoToXY( WhereX, y );
end;

procedure PageUp;
begin
  ResetCurLine;
  if Yoffset > GetYmax then
    Yoffset := Yoffset - 20
  else
    Yoffset := 0;
  WriteEditWindow;
end;

procedure MoveCursor2Target;
begin
  MoveCursorMsg;
  repeat
    inp := readkeyword;
    case inp of
      kbUp   : CursorUp;
      kbDown : CursorDown;
      KbPgUp : PageUp;
      KbPgDn : PageDown;
      kbEnter : ;
    otherwise
      beep;
    end;
    SaveWin( state2 );
    RestoreWin( MenuWin );
    WriteStrAt( 68, 1, '    ', textattr );       {blank out 4 spaces}
    WriteStrAt( 68, 1, int2str( WhereY + Yoffset ), textattr );
    RestoreWin( state2 );
  until inp = kbEnter;
end;

procedure BackSpace;
begin
  if not UpdateFlag then
  begin
    CurLine := BB_ReadLine( WhereY + Yoffset );
    UpdateFlag := true;
  end;
  CursorLeft;
  CurLine[ WhereX + Xoffset ] := ' ';
  WriteCharAt( WhereX, WhereY, 1, ' ', textattr );
end;

procedure Delete;
begin
  if not UpdateFlag then
  begin
    CurLine := BB_ReadLine( WhereY + Yoffset );
    UpdateFlag := true;
  end;
  if ( WhereX + Xoffset ) < length( CurLine ) then
    MoveLeft ( CurLine[ WhereX + Xoffset + 1 ], CurLine[ WhereX + Xoffset ],
               length( CurLine ) - WhereX - Xoffset + 1 )
  else
    CurLine[ WhereX + Xoffset ] := ' ';
  WriteEditWindow;
end;

procedure DeleteToEOL;
var
  i : integer;
begin
  if not UpdateFlag then
  begin
    CurLine := BB_ReadLine( WhereY + Yoffset );
    UpdateFlag := true;
  end;
  for i := ( WhereX + Xoffset ) to CurLine.capacity do
    CurLine[ i ] := ' ';
  for i := 1 to GetXMax do
    WriteCharAt( i, WhereY, 1, CurLine[ Xoffset + i ], textattr );
end;

procedure GoToLine( Linenum: integer );
var
  i: integer;
begin
  if Linenum > BB_Count then
    i := BB_Count + 1
  else
    i := Linenum;
  if i < (( GetYMax div 2 ) + 1 ) then
    Yoffset := 0
  else
    Yoffset := i - (( GetYmax div 2 ) + 1 );
  WriteEditWindow;
  GoToXY( 1, i - Yoffset );
end;

procedure InsertLine;
begin
  BB_Insertline( WhereY + Yoffset, '' );    {inserts empty line in BigBuffer}
  CursorHome;           {sets x to col 1}
  WriteEditWindow;      {update screen}
end;

procedure InsertFile;
var
  s : string(80);
begin
  s := GetNameDialog('Insert File');
  if length( s ) > 0 then
    BB_InsertFile( s, WhereY + Yoffset );
end;

procedure Stats;
var
  i, numlines : integer;
  length    : integer = 0;
  Startline : integer = 0;
begin
  for i := 1 to BB_count do
    if BB_len( i ) > length then
    begin
      length := BB_len( i );
      Startline := i;
      numlines := 1;
    end
    else
      if BB_len( i ) = length then
      inc( numlines );
   StatsMsg( length, Startline, numlines );
   readln;
   WriteEditWindow;
end;

{--------------------------------Submenu items--------------------------------}
{--------------------------------File-----------------------------------------}
procedure FileCmds;
var
  ch : char;
begin
 ch := UpCase( FileSubMenu );
 case ch of
  'I' : InsertFile;
  'P' : PrintFile( filenam );
  'S' : Stats;
  otherwise
    beep;
  end;
  WriteEditWindow;
end;


{--------------------------------Block----------------------------------------}
procedure BlockMark;
var
  inp : word;
  Anchor : integer;
begin
  if ( WhereY + Yoffset ) <= BB_Count then
  begin
    MarkStart := WhereY + Yoffset;
    MarkEnd := MarkStart;
    Anchor := MarkStart;
    repeat
      MarkBlockMsg;
      inp := readkeyword;
      if inp = KbUp then
        CursorUp
      else
        if ( inp = KbDown ) and ((WhereY + Yoffset) < BB_Count ) then
          CursorDown;
      If (WhereY + Yoffset) <= Anchor then
        MarkStart := WhereY + Yoffset;
      If (WhereY + Yoffset) >= Anchor then
        MarkEnd := WhereY + Yoffset;
      WriteEditWindow;
    until inp = KbEnter;
  end
  else
    beep;
end;

procedure BlockCopy;
var
  BlockLen: integer;
begin
  MoveCursor2Target;
  BlockLen := MarkEnd - MarkStart + 1;
  if ((( WhereY + Yoffset ) <= MarkEnd ) and
      (( WhereY + Yoffset ) >= MarkStart )) then
    beep
  else
  begin
    if (( WhereY + Yoffset ) < MarkStart ) then
      BB_CopyBlock( MarkStart + BlockLen, WhereY + Yoffset, BlockLen )
    else
      BB_CopyBlock( MarkStart, WhereY + Yoffset, BlockLen );
    MarkStart := WhereY + Yoffset;
    MarkEnd := MarkStart + BlockLen -1;
    WriteEditWindow;
  end;
end;

procedure BlockUnmark;
begin
  MarkStart := 0;
  MarkEnd := 0;
  WriteEditWindow;
end;

procedure BlockDelete;
begin
  BB_DeleteBlock( MarkStart, MarkEnd );
  if (WhereY + Yoffset) > MarkStart then
      if (WhereY + Yoffset) > MarkEnd then
        GoToLine( WhereY + Yoffset + MarkStart - MarkEnd -1 )
      else
        GoToLine( MarkStart );
  BlockUnmark;
end;

procedure BlockMove;
var
  SaveY,
  BlockLen : integer;
begin
  MoveCursor2Target;
  BlockLen := MarkEnd - MarkStart + 1;
  BB_MoveBlock( MarkStart, WhereY + Yoffset, BlockLen );
  if ( WhereY + Yoffset ) > MarkStart then
    MarkStart := WhereY + Yoffset - BlockLen
  else
    MarkStart := WhereY + Yoffset;
  MarkEnd := MarkStart + BlockLen - 1;
  SaveY := WhereY;
  ClrScr;
  GoToXY( 1,SaveY );
  WriteEditWindow;
end;

procedure BlockShift;
begin
  ShiftMsg;
  repeat
    inp := readkeyword;
    case inp of
      kbLeft : BB_ShiftBlock( MarkStart, MarkEnd, 1, false );
      kbRight: BB_ShiftBlock( MarkStart, MarkEnd, 1, true );
      kbShTab: BB_ShiftBlock( MarkStart, MarkEnd, 4, false );
      kbTab  : BB_ShiftBlock( MarkStart, MarkEnd, 4, true );
      kbEnter: ;
    otherwise
      beep;
    end;
    WriteEditWindow;
  until inp = kbEnter;
end;

procedure BlockWrite;
var
  OutFileName: NameString;
  OutFile    : text;
  i : integer;
begin
  OutFileName := GetNameDialog('Create/Append to File');
  if length( OutFileName ) > 0 then
  begin
    Assign( OutFile, OutFileName );
    Append( OutFile );
    for i := MarkStart to MarkEnd do
      Writeln( OutFile, BB_ReadLine( i ));
    Close( OutFile );
  end;
end;

procedure BlockCmds;
var
  ch : char;
begin
  if MarkStart = 0 then
    BlockMark;
  ch := Upcase( blocksubmenu );
  case ch of
  'B' : begin
          BlockUnmark;
          BlockMark;    {ok}
          end;
  'C' : BlockCopy;    {move cursor first}
  'D' : BlockDelete;
  'M' : BlockMove;    {= copy, delete}
  'S' : BlockShift;
  'U' : BlockUnmark;  {ok}
  'W' : BlockWrite;
  otherwise
    beep;
  end;
  WriteEditWindow;
end;


procedure BlockFix( Index: integer; AddedLine: boolean );
begin
  if AddedLine then
  begin
    if Index < MarkStart then inc( MarkStart );
    if Index < MarkEnd then inc( MarkEnd );
  end
  else
  begin
    if Index < MarkStart then dec( MarkStart );
    if Index < MarkEnd then dec( MarkEnd );
  end;
end;
 

{-----------------------------------Line--------------------------------------}
procedure LineAppend;
var
  tmp : lline;
  indx: integer;
  StartPos : integer = 1;
  LastPos : integer;
begin
  indx := WhereY + Yoffset;
  if( indx > 1 ) and ( indx <= BB_count ) then
  begin
    tmp := BB_Readline( indx );
    BB_DeleteLine( indx );
    if MarkStart > 0 then
      BlockFix( indx, false );
    CursorUp;
    CurLine := BB_ReadLine( WhereY + Yoffset );
    CursorEnd;
    while ( StartPos < length( tmp )) and ( tmp[ StartPos ] = ' ' ) do
      inc( StartPos );
    CurLine := concat( CurLine, copy( tmp, StartPos, length( tmp ) - StartPos + 1));
    BB_WriteLine( WhereY + Yoffset, CurLine );
    WriteEditWindow;
  end
  else
    beep;
end;

procedure LineCopy;
var
  FromIndex, ToIndex : integer;
begin
  FromIndex := WhereY + Yoffset;
  if FromIndex > BB_Count then
    beep
  else
  begin
    MoveCursor2Target;
    ToIndex := WhereY + Yoffset;
    BB_InsertLine( ToIndex, BB_ReadLine( FromIndex ));
    if MarkStart > 0 then
      BlockFix( Toindex, true );
    WriteEditWindow;
  end;
end;

procedure LineDup;
var
  i: integer;
begin
  i := WhereY + Yoffset;
  CursorDown;
  BB_InsertLine( i+1, BB_ReadLine( i ));
  if MarkStart > 0 then
    BlockFix( i, true );
  WriteEditWindow;
end;

procedure LineErase;
begin
  if ( WhereY + Yoffset ) <= BB_Count then
  begin
    ResetCurLine;
    BB_DeleteLine( WhereY + Yoffset );
    if MarkStart > 0 then
      BlockFix( WhereY + Yoffset, false );
    WriteEditWindow;
  end;
  if ( WhereY + Yoffset ) > BB_Count then     {deleted last line}
    CursorUp;
end;

procedure LineMove;
var
  FromIndex, ToIndex : integer;
begin
  FromIndex := WhereY + Yoffset;
  if FromIndex > BB_Count then
    beep
  else
  begin
    MoveCursor2Target;
    ToIndex := WhereY + Yoffset;
    if ToIndex > BB_Count then
      ToIndex := BB_Count;
    BB_MoveLine( FromIndex, ToIndex );
    if MarkStart > 0 then
    begin
      BlockFix( ToIndex, true );
      BlockFix( FromIndex, false );
    end;
    WriteEditWindow;
  end;
end;

procedure LineSplit;
var
  StartPos : integer = 1;
  CursorPos,
  LastPos : integer;
begin
  If not UpdateFlag then
    CurLine := BB_ReadLine( WhereY + Yoffset )
  else
    UpdateFlag := false;                       {"Enter" not needed}
  while ( StartPos < length( CurLine )) and ( CurLine[ StartPos ] = ' ' ) do
    inc( StartPos );
  CursorPos := WhereX + Xoffset;
  LastPos := Trimit( CursorPos - 1 );       {defined first segment}
  {$X+}
  setlength( CurLine, LastPos );
  {$X-}
  BB_WriteLine( WhereY + Yoffset, CurLine );

  CursorDown;
  BB_InsertLine( WhereY + Yoffset, '' );
  LastPos := Trimit( CurLine.capacity );      {end of 2nd segment}
  MoveLeft( CurLine[ CursorPos ], CurLine[ StartPos ], LastPos - CursorPos + 1);
  {$X+}
  setlength( CurLine, LastPos - CursorPos + StartPos );
  {$X-}

  BB_WriteLine( WhereY + Yoffset, CurLine );
  ResetCurLine;
  if MarkStart > 0 then
    BlockFix( Startpos, true );
  WriteEditWindow;
  GoToXY( StartPos, WhereY );
end;

(*---------------------------Search-------------------------------------------
  find outer loop:
    uses start - end from submenu
  find inner loop:
    uses
       c:  case sens/insens;  r: replace;  (y/q dialog) or (y/n/a/q) dialog
-----------------------------------------------------------------------------*)

procedure FindCtrl( fromline, toline: integer );
var
  ch: char;
  i,j : integer;
  AllFlag: boolean = false;
  done: boolean;

  function doline( index: integer ): boolean;
 
    function ckchar( c: char ): boolean;
    begin
      if ( isalphanum( c ) or ( c = '_' ) ) then
        ckchar := false
      else
        ckchar := true;
     end;

  var
    i: integer;
    ok,
    quit: boolean;
  begin
    CurLine := BB_ReadLine( index );
    quit := false;
    i := 1;
    repeat
      {search for pattern match in CurLine}
      repeat
        if CaseFlag then
          i := PosFrom( Pattern, CurLine, i )
        else
          i := PosFrom( Pattern, LoCaseStr( CurLine ), i );
        if WordFlag and ( i > 0 ) then
        {test: if string starts line or char preceding isn't [A..Z,a..z,0..9,_]
         and either string ends line or char after isn't [A..Z,a..z,0..9,_].  }
        begin
          if( ( i = 1 ) or ckchar( char( CurLine[i-1]))
            ) and (
              (( i + length( pattern )) > length( CurLine ))
              or ckchar( char( CurLine[ i + length( pattern )]))
            ) then
            ok := true
          else
          begin
            inc( i );
            ok := false;
          end;
        end
        else
          ok := true;
      until ok;
      if i > 0 then           {found match}
      begin
        inc( found );
        if ( i + length( Pattern )) > GetXMax then
          Xoffset := i + length( Pattern ) - GetXMax
        else
          Xoffset := 0;
        GoToLine( Index );
        ChangeTextAttr( i - Xoffset, WhereY, length( pattern ), color[ 2 ]);
        SaveWin( state2 );
        RestoreWin( MenuWin );
        WriteStrAt( 68, 1, '    ', textattr );       {blank out 4 spaces}
        WriteStrAt( 68, 1, int2str( Index ), textattr ); {line number}
        WriteStrAt( 78, 1, '   ', textattr );        {blank out 3 spaces}
        WriteStrAt( 78, 1, int2str( i ), textattr ); {col number}
        RestoreWin( state2 );
        if ReplaceFlag then
        begin
          if not AllFlag then
          begin
            ch := YesNoAllQuitDialog( 'Replace?');
            if ch = 'a' then
            begin
              AllFlag := true;
              ch := 'y';
            end
            else
              if ch = 'q' then
              begin
                i := 0;
                quit := true;
              end;
          end;
          if ch = 'y' then
          begin
            inc( replaced );
            if length( pattern ) > length( RepPattern ) then
              StringShiftLeft( CurLine, i, length( pattern ) - length( RepPattern ))
            else
              if length( pattern ) < length( RepPattern ) then
                StringShiftRight( CurLine, i, length( RepPattern ) - length( Pattern ));
            move( RepPattern[ 1 ], CurLine[ i ], length( RepPattern ));
            BB_WriteLine( Index, CurLine );
          end;
          inc( i );
        end
        else
        begin
          ch := YesQuitDialog('Again?');
          if ch = 'y' then
            inc( i )
          else
          begin
            i := 0;
            quit := true;
          end;
        end;
      end;
    until (i = 0) or quit;
    doline := quit;
  end;

begin
  done := false;
  if fromline < toline then
  begin
    for i := fromline to toline do
    begin
      if BB_len( i ) >= length( Pattern ) then
        done := doline( i );
      if done then
      begin
        if i = toline then
          startline := 0
        else
        begin
          startline := i;         {saves for search/continue}
          inc( startline );
          endline := toline;
          break;
        end;
      end;
    end;
  end
  else
  begin
    for i := fromline downto toline do
    begin
      if BB_len( i ) >= length( pattern ) then
        done := doline( i );
      if done then
      begin
        if i = toline then
          startline := 0
        else
        begin
          startline := i;
          dec( startline );
          endline := toline;
          break;
        end;
      end;
    end;
  end;
  WriteEditWindow;
  for j:= BB_len( i ) to  CurLine.Capacity do     {clean out garbage}
    CurLine[ j ] := ' ';
end;

{----------------------------main menu items----------------------------------}
procedure LineCmds;
var
  ch : char;
begin
  if ( WhereY + Yoffset ) <= BB_Count then
  begin
      ChangeTextAttr( 1, WhereY, BB_Len( WhereY + Yoffset ) - Xoffset, color[ 4 ] );
    ch := UpCase( LineSubMenu );
    case ch of
      'A' : LineAppend;
      'C' : LineCopy;
      'D' : LineDup;
      'E' : LineErase;
      'I' : InsertLine;
      'M' : LineMove;
      'S' : LineSplit;
    end;
  end
  else
    beep;
  WriteEditWindow;
end;

procedure SearchCmds;
var
  ch : char;
  tempstr : string ( 2 );
begin
  ch := UpCase( SearchSubMenu );
  WriteEditWindow;
  if ( ch = 'C') then
  begin
    if ( StartLine > 0 ) then
      FindCtrl( StartLine, EndLine )
    else
      beep;
  end
  else
    if ( ch = 'R' ) then
    begin
      ResultsMsg( Found, Replaced );
      readln;
    end
    else
    begin
      if (ch = 'F') or (ch = 'U') or (ch = 'D') or (ch = 'B') then
      begin
        found := 0;
        replaced := 0;
        pattern :=  GetNameDialog('Search for');
        CaseFlag := false;
        ReplaceFlag := false;
        WordFlag := false;
        tempstr := OptsDialog;
        i := length( tempstr );
        while( i > 0 ) do
        begin
          if upcase( tempstr[ i ]) = 'C' then
            CaseFlag := true
          else
            if upcase( tempstr[ i ]) = 'R' then
              ReplaceFlag := true
          else
            if upcase( tempstr[ i ]) = 'I' then
              WordFlag := true;
          dec( i );
        end;
        if ReplaceFlag then
          RepPattern := GetNameDialog('Replace with');
        if not CaseFlag then
          Pattern := LoCaseStr( Pattern );
        case ch of
          'F' : FindCtrl( 1, BB_Count );                 {full file search}
          'U' : begin
                  if ( WhereY + Yoffset ) > 1 then
                    FindCtrl( WhereY + Yoffset - 1, 1 )  {cur line-1,  up }
                  else
                    beep;
                end;
          'D' : FindCtrl( WhereY + Yoffset, BB_Count );  {cur line,  down}
          'B' : begin
                  if ( MarkStart = 0 ) then BlockMark;
                  FindCtrl( MarkStart, MarkEnd );          {block}
                end;
        end;
    end
    else
      beep;
  end;
  WriteEditWindow;
end;

{-------------------------control-------------------------------}

procedure Enter;
var
  StartPos : integer;
  i, j, k : integer;
begin
  ResetCurLine;
  Xoffset := 0;
  if InsertMode then
  begin
    StartPos := BB_Start( WhereY + Yoffset );
    CursorDown;
    InsertLine;
    if MarkStart > 0 then
      BlockFix( WhereY + Yoffset, true );
  end
  else
  begin
    CursorDown;
    StartPos := BB_Start( WhereY + Yoffset );
  end;
  if StartPos > GetXMax then
  begin
    Xoffset := StartPos - 1;
    StartPos := 1;
  end;
  WriteEditWindow;
  GoToXY( StartPos, WhereY );
end;


{-------------------------Character handler----------------------------}
procedure HandleCH;
var
 SetColor : TTextAttr;
begin
  if not UpdateFlag then
  begin
    CurLine := BB_ReadLine( WhereY + Yoffset );
    UpdateFlag := true;
  end;
  if (( MarkEnd >= ( WhereY + Yoffset )) and
      ( MarkStart <= ( WhereY + Yoffset ))) then
    SetColor := color[ 1 ]
  else
    SetColor := color[ 3 ];

  { 1. cur pos > length( CurLine ):     setlength, add char to end, cursor
right
    2. cur pos = length( CurLine ):     add char to end, cursor right
    3. cur pos < length( CurLine )
        insert mode:                    add char to end, right shift data,
                                        write char, cursor right
        overwrite mode:                 write char, cursor right }

  if ( WhereX + Xoffset ) > length( CurLine ) then
  begin
    {$X+}
    SetLength( CurLine, ( WhereX + Xoffset ));
    {$X-}
  end
  else
  begin
    if InsertMode then
    begin
      CurLine := Concat( CurLine, ' ' );
      MoveRight( CurLine[ WhereX + Xoffset ], CurLine[ WhereX + Xoffset + 1 ],
                 Length( CurLine ) - WhereX - Xoffset  );
    end;
  end;
  CurLine[ WhereX + Xoffset ] := key2char( inp );

  for i := 1 to GetXMax do
    if  CurLine[ Xoffset + i ] = chTab then
      WriteCharAt( i, WhereY, 1, CurLine[ Xoffset + i ], red )
    else
      WriteCharAt( i, WhereY, 1, CurLine[ Xoffset + i ], SetColor );
  CursorRight;

end;

{----------------------------Main -------------------------------}

begin
  if (ScreenSize.X < MinScreenSizeX ) or (ScreenSize.Y < MinScreenSizeY) then
    writeln('Screen Size is too small (',MinScreenSizeX,'x',MinScreenSizeY,' min)')
  else
  begin
    InitMenuBar;
    InitEditWin;
    ResetCurLine;
    i := BB_getfile;
    if i > 0 then
      GoToLine( i )
    else
    begin
      WriteEditWindow;
      GoToXY( WhereX, 1 );          {start off at 1st line}
    end;
    inp := 0;
    while( inp <> kbAltX ) do
    begin
      i := WhereX + Xoffset;
      j := WhereY + Yoffset;
      SaveWin( state2 );
      RestoreWin( MenuWin );
      if InsertMode then
        WriteStrAt( 56, 1, 'Ins', textattr )
      else
        WriteStrAt( 56, 1, 'Ovl', textattr );
      WriteStrAt( 66, 1, '     ', textattr );      {blank out 5 spaces}
      WriteStrAt( 66, 1, int2str( j ), textattr ); {line number}
      WriteStrAt( 77, 1, '    ', textattr );       {blank out 4 spaces}
      WriteStrAt( 77, 1, int2str( i ), textattr ); {col number}
      RestoreWin( state2 );
      inp := readkeyword;
      case inp of
      {cursor moves}
        kbHome   : CursorHome;
        kbLeft   : CursorLeft;
        kbRight  : CursorRight;
        kbBkSp   : BackSpace;
        kbDel    : Delete;
        kbEnd    : CursorEnd;
        kbShTab : beep;
      {commands}
        kbAltDel  : DeleteToEOL;
        kbAltH    : begin
                      HelpMsg;
                      WriteEditWindow;
                    end;
        kbAltR    : ResetCurLine;
        kbIns     : InsertMode := not( InsertMode );        {toggle}
        kbShIns   : InsertLine;
      {data}
        $20 .. $7F, kbTab : HandleCH;
      otherwise
        begin
          if UpdateFlag then
          begin
            i := Trimit( CurLine.capacity );
            {$X+}
            setlength( CurLine, i );
            {$X-}
            BB_WriteLine( WhereY + Yoffset, CurLine );
            UpdateFlag := false;
          end;
          ResetCurLine;
          case inp of
            kbDown    : CursorDown;
            kbUp      : CursorUp;
            kbPgDn    : PageDown;
            kbPgUp    : PageUp;
            kbEnter   : Enter;
            kbAltB    : BlockCmds;
            kbAltF    : FileCmds;
            kbAltG    : GoToLine( GoToDialog );
            kbAltL    : LineCmds;
            kbAltS    : SearchCmds;
            kbAltX    : break;
          otherwise
            beep;
          end;
        end;
      end;
    end;
    if BB_FileUpdate then
      WriteFileWin( filenam )
    else
      NoFileUpdateMsg;
  end;
end.
