unit peBuffer;  {text buffer & i/o for pascal editor}
{written by Russell Whitaker <russ@ashlandhome.net>, Feb 2000
 vers 0.0.2 June 2001
 0.0.3 Nov 2001 rem getline; mv tab functs to main
 0.1 Jan 2002 changed to var length lines
              released under GPL
 }

interface

const MAXLINES = 20000;

type
  lline         = string ( 4000 );      {buffer for data xfer}

{basic functions}
procedure BB_DeleteLine( Index: integer );
procedure BB_InsertLine( Index: integer; Aline: lline );
procedure BB_MoveLine( FromIndex, ToIndex: integer );
function BB_ReadLine( Index: integer ): lline;
{ procedure BB_UndoLine; note: not implemented }
procedure BB_WriteLine( Index: integer; Aline: lline ); {overwrite existing}

function BB_count : integer;            {# of lines in BigBuffer}
function BB_Len( Index : integer ): integer;   {length of a line}
function BB_Start( Index: integer ): integer;  {first non blank in line}
function BB_fileupdate : boolean;

{block functions}
procedure BB_CopyBlock( BlockStart, ToLine, count: integer );
procedure BB_DeleteBlock( BlockStart, BlockEnd: integer );
procedure BB_MoveBlock( BlockStart, ToLine, count: integer );
procedure BB_ShiftBlock( BlockStart, ToLine, ShiftCount: integer; ShiftRt: boolean );

{file io}
function BB_getfile: integer;    {returns # for goto line}
procedure BB_InsertFile( fileName: string; Index: integer );
procedure BB_Putfile( OutpFileName: string );

var
  filenam : string( 80 );
 
implementation
uses GPCutil,
     StringShift;

type
  ptrline  = ^string;
  TbbArray = array[1..MAXLINES] of ptrline;

var
  BigBuffer  : Tbbarray;                {array of pointers to data}
  LineCount  : integer = 0;             {# of lines in buffer}
  pBlank : ptrline;                     {a universal blank line}
  cap : integer;                        {capacity of line}
  s : string( 80 );
  dummy : boolean;
  tmp : lline;
  over90  : integer = 0;
  fileupdate : boolean = false;

procedure AddLine( Aline: lline ); {used internally}
begin
  inc( LineCount );
  if length( Aline ) > 0 then
  begin
    cap := ( length( Aline ) or 3 ) + 8;
    BigBuffer[ LineCount ] := new( ptrline, cap );
    BigBuffer[ LineCount ]^ := Aline;
  end
  else
    BigBuffer[ LineCount ] := pBlank;
end;

procedure BB_DeleteLine( Index: integer );
var
  i : integer;
begin
  i := Index;
  if BigBuffer[ i ] <> pBlank then
    dispose( BigBuffer[ i ] );
  while i < LineCount do
  begin
    BigBuffer[ i ] := BigBuffer[ i + 1 ];
    inc( i );
  end;
  BigBuffer[ LineCount ] := nil;
  dec( LineCount );
  fileupdate := true;
end;


Procedure BB_InsertLine( Index: integer; Aline: lline );
var
  i: integer;
begin
  if Index <= LineCount then            {make a hole}
    for i := LineCount downto Index  do
      BigBuffer[ i + 1 ] := BigBuffer[ i ];
  if length( Aline ) > 0 then
  begin
    cap := ( length( Aline ) or 3 ) + 4;
    BigBuffer[ Index ] := new( ptrline, cap );
    BigBuffer[ Index ]^ := Aline;
  end
  else
    BigBuffer[ Index ] := pBlank;
  inc( LineCount );
  fileupdate := true;
end;

procedure BB_MoveLine( FromIndex, ToIndex: integer );
var
  i : integer;
  temp: ptrline;
begin
  temp := BigBuffer[ FromIndex ];
  if FromIndex < ToIndex then           { shift lines down}
    for i := FromIndex to ToIndex - 1 do
      BigBuffer[ i ] := BigBuffer[ i + 1 ]
  else                                  { shift lines up}
    for i := FromIndex downto ToIndex + 1 do
      BigBuffer[ i ] := BigBuffer[ i - 1 ];
  BigBuffer[ ToIndex ] := temp;
  fileupdate := true;
end;

function BB_ReadLine( Index: integer ): lline;
var
  temp : string( 1 ) = '';
begin
  if Index <= LineCount then
    BB_ReadLine := BigBuffer[ Index ]^
  else
  begin
    BB_ReadLine := temp;
  end;
end;

procedure BB_WriteLine( Index: integer; Aline: lline );
begin
  if Index > LineCount then
    Addline( Aline )
  else
  begin
    if length( Aline ) > BigBuffer[ Index ]^.CAPACITY then
    begin
      if BigBuffer[ Index ] <> pBlank then
        dispose( BigBuffer[ Index ] );
      cap := ( length( Aline ) or 3 ) + 4;
      BigBuffer[ Index ] := new( ptrline, cap );
    end;
    {now capacity is adaquate or Aline is 0}
    if length( Aline ) <> 0 then
      BigBuffer[ Index ]^ := Aline
    else
      if BigBuffer[ Index ] <> pBlank then
      begin
        dispose( BigBuffer[ Index ] );
        BigBuffer[ Index ] := pBlank;
      end;
  end;
  fileupdate := true;
end;

function BB_Count: integer;
begin
  BB_Count := LineCount;
end;

function BB_Len( Index : integer ): integer;
begin
  BB_Len := Length( BigBuffer[ Index ]^ );
end;

function BB_Start( Index: integer ): integer;  {first non blank}
var
  i: integer = 1;
begin
  if Index <= LineCount then
  begin
    while(( BigBuffer[Index]^[i] = ' ')
          and ( i < Length( BigBuffer[ Index ]^))) do
      inc( i );
    BB_Start := i;
  end
  else
    BB_Start := 1;
end;

function BB_fileupdate: boolean;
begin
  BB_fileupdate := fileupdate;
end;

{-----------------block functions--------------------------}
procedure BB_CopyBlock( BlockStart, ToLine, count: integer );
var
  i,j : integer;
begin
  j := LineCount + count;
  for i := LineCount downto ToLine do  { make hole }
  begin
    BigBuffer[ j ] := BigBuffer[ i ];
    dec( j );
  end;
  LineCount := LineCount + count;
{BlockStart is where the block is after the hole is made}
  j := BlockStart;
  for i := ToLine to ToLine + count - 1 do
  begin
    BigBuffer[ i ] := new( ptrline, length( BigBuffer[ j ]^ ));
    BigBuffer[ i ]^ := BigBuffer[ j ]^;
    inc( j );
  end;
  fileupdate := true;
end;

procedure BB_DeleteBlock( BlockStart, BlockEnd: integer );
var
  i,j : integer;
begin
  for i := BlockStart to BlockEnd do    {del block}
    if BigBuffer[ i ] <> pBlank then
      dispose( BigBuffer[ i ] );
  i := BlockStart;
  j := BlockEnd + 1;
  while j <= LineCount do               {fill hole}
  begin
    BigBuffer[ i ] := BigBuffer[ j ];
    BigBuffer[ j ] := nil;
    inc( i );
    inc( j );
  end;
  LineCount := LineCount - BlockEnd + BlockStart - 1;
  fileupdate := true;
end;

procedure BB_MoveBlock( BlockStart, ToLine, count: integer );
var
  i,j : integer;
  temp : array[ 1..count ] of ptrline;
begin
  {first move the block to temp}
  for i := 1 to count do
    temp[ i ] := BigBuffer[ BlockStart -1 + i ];

  {next shift the middle}
  j := BlockStart;
  if BlockStart < ToLine then           {shift down, up loop}
  begin
    for i := BlockStart + count to ToLine - 1 do
    begin
      BigBuffer[ j ] := BigBuffer[ i ];
      inc( j );
    end;
  end
  else                                  {shift up, downto loop}
  begin
    for i := BlockStart + count - 1 downto ToLine + count do
    begin
      dec( j );
      BigBuffer[ i ] := BigBuffer[ j ];
    end;
  end;

  {move temp into hole}
  for i := 1 to count  do
  begin
    BigBuffer[ j ] := temp[ i ];
    inc( j );
  end;
  fileupdate := true;
end;

procedure BB_ShiftBlock( BlockStart, ToLine, ShiftCount: integer; ShiftRt: boolean );
var
  i,j : integer;
  Aline : lline;
begin
  for i := BlockStart to ToLine do
  if ShiftRt then
  begin
    if Length( BigBuffer[ i ]^) > 0 then    {line contains chars}
    begin
      {check if need added capacity}
      if (length( BigBuffer[ i ]^) + ShiftCount) >= BigBuffer[ i ]^.CAPACITY then
      begin
        Aline := BigBuffer[ i ]^;
        dispose( BigBuffer[ i ] );
        cap := ( length( Aline ) or 3 ) + 8;
        BigBuffer[ i ] := new( ptrline, cap );
        BigBuffer[ i ]^ := Aline;
      end;
      {make hole ShiftCount wide}
      StringShiftRight( BigBuffer[ i ]^, 1, ShiftCount );
      for j := 1 to ShiftCount do           {fill hole}
        BigBuffer[ i ]^[ j ] := ' ';
    end;
  end
  else
  begin
    {close hole ShifCount wide}
    StringShiftLeft( BigBuffer[ i ]^, 1, ShiftCount );
  end;
  fileupdate := true;
end;

{-----------------file io------------------------------------------}
function ReadFile( Afilename: string): integer;  {return error code}
var
  Infile: text;
  errorcode : integer;
begin
  assign( InFile, Afilename );
  {$I-}
  reset( InFile );
  {$I+}
  errorcode := IoResult;
  if errorcode = 0 then
  begin
    while not EOF( InFile ) do
    begin
      Readln( InFile, tmp );
      if length( tmp ) > 90 then inc( over90 );
      AddLine( tmp );
    end;
    close( InFile );
  end;
  ReadFile := errorcode;
end;

function BB_getfile: integer;
var
  i : integer = 0;
begin
  if ParamCount >= 1 then               {first need a file name}
    filenam := ParamStr( 1 )
  else
  begin
    writeln('Input name of file to edit: ');
    readln( s );
    if length( s ) > 0 then
      filenam := s
    else
      filenam := 'untitled.pas';
  end;
  i := ReadFile( filenam );                {next try to read it}
  if i = 0 then
  begin
    if ParamCount = 2 then             {no goto: exit 1A}
    begin
      dummy := str2int( ParamStr( 2 ), i );
      BB_getfile := i;
    end
    else                               {usual condition: exit 1B}
    begin
      writeln('Editing file: ', filenam );
      writeln('# of lines is ', BB_count );
      writeln('# of lines over 90 = ', over90 );
      writeln('Hit Enter to continue');
      readln;
      BB_getfile := 0;
    end;
  end
  else                                  {no file loaded: exit 2}
  begin
    writeln('Unable to open: ', filenam );
    writeln('Assuming new file');
    writeln('Hit Enter to continue');
    readln;
    BB_getfile := 0;
  end;
end;

procedure BB_InsertFile( fileName: string; Index: integer );
var
  i : integer;
begin
  i := LineCount;
  ReadFile( fileName );
  If ( Index < i ) and ( i < LineCount ) then
    BB_MoveBlock( i+1, index, LineCount - i );
  fileupdate := true;
end;

procedure BB_Putfile( OutpFileName: string );
var
  Outfile: text;
  i : integer;
begin
  If LineCount > 0 then
  begin
    assign( Outfile, OutpFileName );
    rewrite( Outfile );
    while ( LineCount > 1 ) and then ( BigBuffer[ LineCount ] = pBlank ) do
      dec( LineCount );
    for i := 1 to LineCount do
    begin
      writeln( Outfile, BigBuffer[ i ]^);
      if BigBuffer[ i ] <> pBlank then
        dispose( BigBuffer[ i ] );
    end;
    close( Outfile );
    fileupdate := false;
  end;
end;

to begin do
begin
  pBlank := new( ptrline, 0 );
  pBlank^ := '';      {just in case}
end;

to end do
  dispose( pBlank );

end.
