Unit peMenu; {screen routines for pascal editor}
{written by Russell Whitaker <russ@ashlandhome.net>, Feb 2000
 updated July 2001
 10/17/2002: changed YesNoDialog to YesQuitDialog and YesNoAllQuitDialog
 11/11/2002: added continue to SearchSubMenu
  1/20/2003: added "w" to OptsDialog
  3/4/2003: changed "w" to "i" }
interface
uses GPC, CRT;

type
  NameString = string ( 80 );

  procedure InitMenuBar;
  procedure InitEditWin;

  function FileSubMenu: char;
  function BlockSubMenu: char;
  function LineSubMenu: char;
  function SearchSubMenu: char;
  function GetNameDialog( const Title: string ): NameString;
  function GoToDialog: integer;
  function OptsDialog: NameString;
  function YesQuitDialog( const Title: string ): char;
  function YesNoAllQuitDialog( const Title: string ): char;

  procedure MarkBlockMsg;
  procedure ShiftMsg;
  procedure MoveCursorMsg;
  procedure NoFileUpdateMsg;
  procedure ResultsMsg(Found, Replaced: integer );
  procedure StatsMsg(Length, Start, count: integer );
  procedure WriteFileWin( AfileName: string );
  procedure HelpMsg;

var
  MenuWin, EditWin, SubMenu : winstate;
  CurSize : Tpoint;

type
  TColorPallet = array[ 1 .. 5 ] of Ttextattr;

const
  Color : TColorPallet = (( 16 * LightGray) + Black ,( 16 * LightGray ) + Red,
                          ( 16 * Blue ) + LightGray, ( 16 * Blue ) + Yellow,
                          ( 16 * LightGray) + Magenta );
implementation

uses GpcUtil, FrameWin, pebuffer;

const
  HelpText : array[1..18] of string( 48 ) = (
    'Command Line Options:',
    '  pe <File Name> <go to line number>',
    'Activate Main Menu items with ALT-letter',
    '  File submenu',
    '  Block submenu',
    '    If block is not marked then ALT-b starts',
    '    mark routine followed by block submenu',
    '  Line submenu',
    '  Search submenu',
    '  Help: one page',
    '  eXit: also asks exit options',
    '  Insert key toggles INSert/OVerlay modes',
    'Cursor moves:',
    '  Cursor up dn left right; page up dn; Home, End',
    'Hot keys Not on menu:',
    '  ALT-delete - deletes from cursor to EOL',
    '  ALT-R - restores current line from memory',
    '  Sh-Insert - insert new line at cursor');

  HelpTextColor : array[1..18] of integer =
    ( 4,3,4,3,3,3,3,3,3,3,3,3,4,3,4,3,3,3 );
 
{menu bar at top of screen}
procedure InitMenuBar;
begin
  makewin( MenuWin, 1, 1, CurSize.X, 1 );
  fillwin( '.', Color[ 1 ] );
  WriteStrAt( 1, 1, ' Pascal Editor', Color[ 2 ] );
  textbackground( lightgray );
  textcolor( black);
  writestrat( 17, 1, 'F', color[ 5 ] );
  writestrat( 18, 1, 'ile  ', color[ 1 ] );
  writestrat( 23, 1, 'B', color[ 5 ] );
  writestrat( 24, 1, 'lock  ', color[ 1 ] );
  writestrat( 30, 1, 'L', color[ 5 ] );
  writestrat( 31, 1, 'ine   ', color[ 1 ] );
  writestrat( 36, 1, 'S', color[ 5 ] );
  writestrat( 37, 1, 'earch  ', color[ 1 ] );
  writestrat( 44, 1, 'H', color[ 5 ] );
  writestrat( 45, 1, 'elp  e', color[ 1 ] );
  writestrat( 51, 1, 'X', color[ 5 ] );
  writestrat( 52, 1, 'it   ', color[ 1 ] );
  writestrat( 60, 1, ' line ', color[ 1 ] );
  writestrat( 73, 1, 'col ', color[ 1 ] );
  savewin( MenuWin );
end;

procedure InitEditWin;
begin
  textbackground( blue );
  textcolor( lightgray );
  makewin( EditWin, 1, 2, CurSize.X, CurSize.Y );
  fillwin( ' ', textattr );
end;

function FileSubMenu: char;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  HideCursor;
  makewin( SubMenu, 62, 3, 76, 8 );
  FrameWin('File Menu', SingleFrame );
  fillwin( ' ', textattr );
  writestring( 'Insert',  1, 2 );
  writestring( 'Print ',  2, 2 );
  writestring( 'Stats ',  3, 2 );
{ writestring( '     ',      4, 2 );
  writestring( '     ',      5, 2 );}
  FileSubMenu := char( readkeyword ); {avoids cursor readkey problem}
  RestoreWin( EditWin );
end;

function BlockSubMenu: char;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  HideCursor;
  makewin( SubMenu, 62, 3, 76, 11 );
  FrameWin('Block Menu', SingleFrame );
  fillwin( ' ', textattr );
  writestring( 'mark Block',  1, 2 );
  writestring( 'Unmark ',     2, 2 );
  writestring( 'Copy',        3, 2 );
  writestring( 'Delete',      4, 2 );
  writestring( 'Move',        5, 2 );
  writestring( 'Shift',       6, 2 );
  writestring( 'Write file',  7, 2 );
  BlockSubMenu := char( readkeyword ); {avoids cursor readkey problem}
  RestoreWin( EditWin );
end;

function LineSubMenu: char;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  HideCursor;
  makewin( SubMenu, 62, 3, 76, 11 );
  FrameWin('Line Menu', SingleFrame );
  fillwin( ' ', textattr );
  writestring( 'Append',    1, 2 );
  writestring( 'Copy',      2, 2 );
  writestring( 'Duplicate', 3, 2 );
  writestring( 'Erase',     4, 2 );
  writestring( 'Insert',    5, 2 );
  writestring( 'Move',      6, 2 );
  writestring( 'Split',     7, 2 );
  LineSubMenu := char( readkeyword );
  RestoreWin( EditWin );
end;

function SearchSubMenu: char;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  HideCursor;
  makewin( SubMenu, 62, 3, 77, 11 );
  FrameWin('Search Menu', SingleFrame );
  fillwin( ' ', textattr );
  writestring( 'Full',       1, 2 );
  writestring( 'Up',         2, 2 );
  writestring( 'Down',       3, 2 );
  writestring( 'Block',      4, 2 );
  writecharat( 1,5,14, '-', color[ 1 ] );
  writestring( 'Continue',   6, 2 );
  writestring( 'Results',    7, 2 );
  SearchSubMenu := char( readkeyword );
  RestoreWin( EditWin );
end;

function GetNameDialog( const Title: string ): NameString;
var
  s : NameString;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 32, 3, 74, 5 );
  FrameWin( Title, SingleFrame );
  fillwin( ' ', color[ 3 ] );
  readln( s );
  GetNameDialog := s;
  RestoreWin( EditWin );
end;

function GetNumber : integer;      {internal function}
var
  ch: char;
  i : integer = 0;
begin
  repeat
    ch := readkey;
    if ( ch >= '0' ) and ( ch <= '9' ) then
    begin
      i := ( 10 * i ) + integer( ch ) - 48;
      Write( ch );
    end;
  until ch = chEnter;
  GetNumber := i;
end;

function GoToDialog : integer;
var
  i : integer = 0;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 62, 3, 74, 5 );
  FrameWin('Go To Line', SingleFrame );
  fillwin( ' ', color[ 3 ] );
  GoToDialog := GetNumber;
  RestoreWin( EditWin );
end;

function OptsDialog: NameString;
var
  s : NameString;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 62, 9, 75, 14 );
  FrameWin('Options', SingleFrame );
  fillwin( ' ', color[ 3 ] );
  writestrat( 1, 1, ' C', color[ 5 ] );
  writestrat( 3, 1, 'heck Case ', color[ 1 ] );
  writestrat( 1, 2, ' I', color[ 5 ] );
  writestrat( 3, 2, 'dentifier ', color[ 1 ] );
  writestrat( 1, 3, ' R', color[ 5 ] );
  writestrat( 3, 3, 'eplace    ', color[ 1 ] );
  GoToXY( 1,4 );
  readln( s );
  RestoreWin( EditWin );
  OptsDialog := s;
end;

function YesQuitDialog( const Title: string ): char;
var
  ch : char;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 64, GetYMax - 4, 76, GetYMax - 2 );
  FrameWin( Title, SingleFrame );
  fillwin( ' ', color[ 3 ] );
  write( 'Y/Q ' );
  repeat
    ch := LoCase( ReadKey );
  until (ch = 'y') or (ch = 'q') or (ch = 'n');
  RestoreWin( EditWin );
  YesQuitDialog := ch;
end;

function YesNoAllQuitDialog( const Title: string ): char;
var
  ch : char;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 60, GetYMax - 4, 76, GetYMax - 2 );
  FrameWin( Title, SingleFrame );
  fillwin( ' ', color[ 3 ] );
  write( 'Y/N/A/Q ' );
  repeat
    ch := LoCase( ReadKey );
  until (ch = 'y') or (ch = 'n') or (ch = 'a' ) or (ch = 'q');
  RestoreWin( EditWin );
  YesNoAllQuitDialog := ch;
end;

procedure MarkBlockMsg;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 42, 3, 66, 8 );
  FrameWin('Marking Block', DoubleFrame );
  fillwin( ' ', color[ 3 ] );
  WriteStrAt( 2,2, 'Use Up and Down Cursor keys', color[ 4 ]);
  WriteStrAt( 2,3, '    Enter when done', color[ 4 ]);
  RestoreWin( EditWin );
end;

procedure ShiftMsg;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 42, 3, 74, 8 );
  FrameWin('Shifting Block', DoubleFrame );
  fillwin( ' ', color[ 3 ] );
  WriteStrAt( 2,1, 'Cursor Right/Left keys: 1 col' , color[ 4 ]);
  WriteStrAt( 2,2, 'Tab/ShiftTab: move 4 columns' , color[ 4 ]);
  WriteStrAt( 2,4, '    Enter when done', color[ 4 ]);
  RestoreWin( EditWin );
end;

procedure MoveCursorMsg;
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 42, 3, 66, 8 );
  FrameWin('Information', DoubleFrame );
  fillwin( ' ', color[ 3 ] );
  WriteStrAt( 2,2, 'Move Cursor to target', color[ 4 ]);
  WriteStrAt( 2,3, 'line and press Enter', color[ 4 ]);
  RestoreWin( EditWin );
end;

procedure NoFileUpdateMsg;
begin
  makewin( SubMenu, 42, 3, 66, 8 );
  FrameWin('Information', DoubleFrame );
  fillwin( ' ', color[ 3 ] );
  WriteStrAt( 2,2, ' File not changed --', color[ 4 ]);
  WriteStrAt( 2,3, ' so no update needed', color[ 4 ]);
end;

procedure ResultsMsg(Found, Replaced: integer );
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 51, 3, 76, 9 );
  FrameWin('Search Statistics', DoubleFrame );
  fillwin( ' ', color[ 3 ] );
  WriteStrAt( 3,2, 'Found:     ' + Int2Str( Found ), color[ 4 ]);
  WriteStrAt( 3,3, 'Replaced:  ' + Int2Str( Replaced ), color[ 4 ]);
  WriteStrAt( 4,5, 'Hit Enter when done', color[ 4 ]);
  RestoreWin( EditWin );
end;

procedure StatsMsg(length, start, count: integer );
begin
  SaveWin( EditWin );
  textbackground( lightgray );
  textcolor( Black );
  makewin( SubMenu, 51, 3, 76, 10 );
  FrameWin('File Statistics', DoubleFrame );
  fillwin( ' ', color[ 3 ] );
  WriteStrAt( 3,2, 'Longest line:   ' + Int2Str( Length ), color[ 4 ]);
  WriteStrAt( 3,3, 'First occurs:   ' + Int2Str( Start ), color[ 4 ]);
  WriteStrAt( 3,4, '      Occurs:   ' + Int2Str( count ), color[ 4 ]);
  WriteStrAt( 4,6, 'Hit Enter when done', color[ 4 ]);
  RestoreWin( EditWin );
end;

procedure WriteFileWin( AFileName: string );
var
  i : integer = 0;
  inp : integer;
  tempStr : string( 80 );
  F : text;
begin
  makewin( submenu, 2, 3, 50, 12);
  FrameWin('Write File Dialog', SingleFrame );
  fillwin( ' ', color[ 3 ] );
  WriteStrAt( 2,2, 'Save changes to ' + AFileName + '?', color[ 4 ]);
  WriteStrAt( 2,3, 'N = no, C = change name, Any other key = ok', color[ 3 ]);
  GoToXY( 2, 4 );
  inp := readkeyword;
  If not ( upcase( char( inp )) = 'N' ) then
  begin
    if upcase( char( inp )) = 'C' then
    begin
      WriteStrAt( 2,4, 'Input new file name:', color[ 4 ]);
      GoToXY( 2, 5);
      readln( tempStr );
    end
    else
      tempStr := AFileName;
    writeln;
    if FileExists( tempStr ) then
    begin
      if FileExists( tempStr + '~' ) then
      begin
        Assign( F, tempStr + '~' );
        Erase( F );
        Writeln( '   Deleted old backup ');
      end;
      Assign( F, tempStr );
      Rename( F, tempStr + '~' );
      Writeln( '   Made new backup ');
    end;
    BB_PutFile( tempStr );
    Writeln( '   Wrote file' );
  end
  else
  begin
    Writeln;
    Writeln( '   Changes not saved ');
  end;
end;

procedure HelpMsg;
var
  i : integer;
begin
  i := (CurSize.X - 52) div 2;
  SaveWin( EditWin );
  makewin( SubMenu, i, 2, i + 52, 22 );
  FrameWin('HELP: hit Enter to exit', DoubleFrame );
  fillwin( ' ', color[ 3 ] );
  HideCursor;
  for i := 1 to 19 do
    WriteStrAt( 2,i, HelpText[ i ], color[ HelpTextColor[ i ]] );
  readln;
  RestoreWin( EditWin );
end;

To begin do
    CurSize := ScreenSize;

end.
